@@ -2,7 +2,7 @@
 # -*- coding: utf-8 -*-
 """Exosite RPC API Command Line Interface
    Provides command line access to the Remote Procedure Call API:
-   https://github.com/exosite/api/tree/master/rpc
+   https://github.com/exosite/docs/tree/master/rpc
 
 Usage:
   exo [--help] [options] <command> [<args> ...]
@@ -15,6 +15,7 @@ Options:
   --httptimeout=<sec>  HTTP timeout [default: 60]
   --https              Enable HTTPS (deprecated, HTTPS is default)
   --http               Disable HTTPS
+  --useragent=<ua>     Set User-Agent Header for outgoing requests
   --debug              Show debug info (stack traces on exceptions)
   -d --debughttp       Turn on debug level logging in pyonep
   --discreet           Obfuscate RIDs in stdout and stderr
@@ -31,8 +32,13 @@ See 'exo <command> --help' for more information on a specific command.
 from __future__ import unicode_literals
 import sys
 import os
+from os.path import expanduser
 import json
-import csv
+if sys.version_info < (3, 0):
+    import unicodecsv as csv
+else:
+    import csv
+import platform
 import re
 from datetime import datetime
 from datetime import timedelta
@@ -116,14 +122,15 @@ Command options:
     exo [options] record <cik> [<rid>] ((--value=<timestamp,value> ...) | -)
     exo [options] record <cik> [<rid>] --interval=<seconds> ((--value=<value> ...) | -)
 
-    - reads data from stdin.
+    - reads data from stdin. Data should be in CSV format (no headers) with rows
+      like this: <unix timestamp>,<value>
     --interval generates timestamps at a regular interval into the past.'''),
     ('create',
         '''Create a resource from a json description passed on stdin (with -),
     or using command line shorthand (other variants).\n\nUsage:
     exo [options] create <cik> (--type=client|clone|dataport|datarule|dispatch) -
     exo [options] create <cik> --type=client
-    exo [options] create <cik> --type=dataport (--format=binary|boolean|float|integer|string)
+    exo [options] create <cik> --type=dataport (--format=float|integer|string)
 
 Command options:
     --name=<name     set a resource name (overwriting the one in stdin if present)
@@ -135,7 +142,9 @@ Command options:
 Details:
     Pass - and a json description object on stdin, or leave it off to use defaults.
     Description is documented here:
-    http://developers.exosite.com/display/OP/Remote+Procedure+Call+API#RemoteProcedureCallAPI-create
+    https://github.com/exosite/docs/tree/master/rpc#create-client
+    https://github.com/exosite/docs/tree/master/rpc#create-dataport
+    https://github.com/exosite/docs/tree/master/rpc#create-datarule
 
     If - is not present, creates a resource with common defaults.'''),
     ('listing',
@@ -164,6 +173,7 @@ Command options:
     --cikonly      print CIK by itself
     --pretty       pretty print output
     --recursive    embed info for any children recursively
+    --level=<num>  number of levels to recurse through the client tree
     --include=<key list>
     --exclude=<key list>
                    comma separated list of info keys to include and exclude.
@@ -172,9 +182,9 @@ Command options:
                    all available keys are returned.'''),
     ('update',
         '''Update a resource from a json description passed on stdin.\n\nUsage:
-    exo [options] update <cik> (<rid> - | -)
+    exo [options] update <cik> <rid> -
 
-    For details see https://github.com/exosite/api/tree/master/rpc#update'''),
+    For details see https://github.com/exosite/docs/tree/master/rpc#update'''),
     ('map',
         '''Add an alias to a resource.\n\nUsage:
     exo [options] map <cik> <rid> <alias>'''),
@@ -203,8 +213,14 @@ Command options:
     --all-children  drop all children of the resource.
     {{ helpoption }}'''),
     ('flush',
-        '''Remove all time series data from a resource.\n\nUsage:
-    exo [options] flush <cik> [<rid>]'''),
+        '''Remove time series data from a resource.\n\nUsage:
+    exo [options] flush <cik> [<rid>]
+
+Command options:
+    --start=<time>  flush all points newer than <time> (exclusive)
+    --end=<time>    flush all points older than <time> (exclusive)
+
+    If --start and --end are both omitted, all points are flushed.'''),
     ('usage',
         '''Display usage of One Platform resources over a time period.\n\nUsage:
     exo [options] usage <cik> [<rid>] --start=<time> [--end=<time>]
@@ -286,6 +302,23 @@ Command options:
     exo [options] activate <cik> (--client=<child-cik> | --share=<code>)'''),
     ('deactivate', '''Deactivate a share code or expire a CIK\n\nUsage:
     exo [options] deactivate <cik> (--client=<cik> | --share=<code>)'''),
+    ('clone', '''Create a clone of a client\n\nUsage:
+    exo [options] clone <cik> (--rid=<rid> | --share=<code>)
+
+Command options:
+     --noaliases     don't copy aliases
+     --nohistorical  don't copy time series data
+     --noactivate    don't activate CIK of clone (client only)
+
+     The clone command copies the client resource specified by --rid or --share
+     into the client specified by <cik>. For example, to clone a portals device,
+     pass the portal CIK as <cik> and the device RID as <rid>.
+
+     The clone and copy commands do similar things, but clone uses the RPC's
+     create (clone) functionality, which is more full featured.
+     https://github.com/exosite/docs/tree/master/rpc#create-clone
+
+     Use the clone command except if you need to copy a device to another portal.'''),
     #('tag', '''Add a tag to a resource\n\nUsage:
     #exo [options] tag <cik> [<rid> ...] [--add=<tag1,tag2>]'''),
     ])
@@ -305,28 +338,43 @@ doc_replace = {
     '{{ helpoption }}': '''    -h --help  Show this screen.''',
 }
 
-# load plugins. use timezone because this file may be running
-# as a script in some other location.
-plugin_path = os.path.join(os.path.dirname(timezone.__file__), 'plugins')
-
-plugin_names = [os.path.basename(f)[:-3]
-    for f in glob.glob(plugin_path + "/*.py")
-    if not os.path.basename(f).startswith('_')]
-
 plugins = []
-for module_name in plugin_names:
+if platform.system() != 'Windows':
+    # load plugins. use timezone because this file may be running
+    # as a script in some other location.
+    plugin_path = os.path.join(os.path.dirname(timezone.__file__), 'plugins')
+
+    plugin_names = [os.path.basename(f)[:-3]
+        for f in glob.glob(plugin_path + "/*.py")
+        if not os.path.basename(f).startswith('_')]
+
+    for module_name in plugin_names:
+        try:
+            plugin = importlib.import_module('plugins.' + module_name)
+        except:
+            plugin = importlib.import_module('exoline.plugins.' + module_name, package='test')
+
+        # instantiate plugin
+        p = plugin.Plugin()
+        plugins.append(p)
+
+        # get documentation
+        cmd_doc[p.command()] = plugin.__doc__
+else:
     try:
-        plugin = importlib.import_module('plugins.' + module_name)
-    except:
-        plugin = importlib.import_module('exoline.plugins.' + module_name, package='test')
-
-    # instantiate plugin
-    p = plugin.Plugin()
-    plugins.append(p)
-
-    # get documentation
-    cmd_doc[p.command()] = plugin.__doc__
-
+        try:
+            from ..exoline.plugins import spec
+        except:
+            from exoline.plugins import spec
+        p = spec.Plugin()
+        plugins.append(p)
+        cmd_doc[p.command()] = spec.__doc__
+    except Exception as ex:
+        import traceback
+        traceback.print_exc()
+        print("Couldn't import spec......")
+        pprint(ex)
+        print('yup, that was an exception.')
 
 # perform substitutions on command documentation
 for k in cmd_doc:
@@ -347,7 +395,7 @@ class ExolineOnepV1(onep.OnepV1):
 class ExoRPC():
     '''Wrapper for pyonep RPC API.
     Raises exceptions on error and provides some reasonable defaults.'''
-    regex_rid = re.compile("[0-9a-zA-Z]{40}")
+    regex_rid = re.compile("[0-9a-fA-F]{40}")
 
     class RPCException(Exception):
         pass
@@ -358,15 +406,18 @@ class ExoRPC():
                  httptimeout=60,
                  https=False,
                  verbose=True,
-                 logrequests=False):
+                 logrequests=False,
+                 user_agent=None):
 
         if port is None:
             port = DEFAULT_PORT_HTTPS if https else DEFAULT_PORT
+        if user_agent is None:
+            user_agent = "Exoline {0}".format(__version__)
         self.exo = ExolineOnepV1(host=host,
                                port=port,
                                httptimeout=httptimeout,
                                https=https,
-                               agent="Exoline {0}".format(__version__),
+                               agent=user_agent,
                                reuseconnection=True,
                                logrequests=logrequests)
 
@@ -384,7 +435,6 @@ class ExoRPC():
            messages.'''
         self._raise_for_response(isok, response)
         if type(response) is list:
-            # TODO: does this always indicate an error condition?
             raise ExoRPC.RPCException(', '.join(['{0}: {1}'.format(msg, t) for msg, t in response]))
 
     def _raise_for_deferred(self, responses):
@@ -425,7 +475,7 @@ class ExoRPC():
             if it doesn't look like a CIK.'''
         if self.regex_rid.match(cik) is None:
             # if cik doesn't look like a cik, maybe it's a shortcut
-            configfile = os.path.join(os.environ['HOME'], '.exoline')
+            configfile = os.path.join(expanduser('~'), '.exoline')
             try:
                 with open(configfile) as f:
                     config = yaml.safe_load(f)
@@ -682,57 +732,59 @@ class ExoRPC():
         self._raise_for_response(isok, response)
         return response
 
-    def listing(self, cik, types, options=None):
-        if options is None:
-            # TODO: remove all uses of this deprecated listing variant
-            isok, response = self.exo.listing(cik, types)
-        else:
-            isok, response = self.exo.listing(cik, types, options)
+    def listing(self, cik, types, options={}):
+        isok, response = self.exo.listing(cik, types, options)
         self._raise_for_response(isok, response)
         return response
 
-    def _listing_with_info(self, cik, types, options={}):
-        '''Return a list of dicts for each type containing mappings between rids and
-        info for that RID. E.g.:
-        [{'<rid0>':<info0>, '<rid1>':<info1>}, {'<rid2>':<info2>}, [], {'<rid3>': <info3>}]'''
+    def _listing_with_info(self, cik, types, info_options={}):
+        '''Return a dict mapping types to dicts mapping RID to info for that
+        RID. E.g.:
+            {'client': {'<rid0>':<info0>, '<rid1>':<info1>},
+             'dataport': {'<rid2>':<info2>, '<rid3>':<info3>}}'''
 
         assert(len(types) > 0)
 
-        listing = self._exomult(cik,
-                                [['listing', types]])[0]
+        listing = self._exomult(cik, [['listing', types, {}]])[0]
 
-        # listing is a list of lists per type, like: [['<rid0>', '<rid1>'], ['<rid2>'], [], ['<rid3>']]
+        # listing is a dictionary mapping types to lists of RIDs, like this:
+        # {'client': ['<rid0>', '<rid1>'], 'dataport': ['<rid2>', '<rid3>']}
 
         # request info for each rid
         # (rids is a flattened version of listing)
-        rids = list(itertools.chain(*listing))
-        responses = self._exomult(cik, [['info', rid, options] for rid in rids])
+        rids = []
+        for typ in types:
+            rids += listing[typ]
 
-        # From the return values make a list of dicts like this:
-        # [{'<rid0>':<info0>, '<rid1>':<info1>}, {'<rid2>':<info2>}, [], {'<rid3>': <info3>}]
-        # use ordered dicts in case someone cares about order in the output (?)
+        responses = self._exomult(cik, [['info', rid, info_options] for rid in rids])
+
+        # From the return values make a dict of dicts
+        # use ordered dicts in case someone cares about order in the output
         response_index = 0
-        listing_with_info = []
-        for type_list in listing:
+        listing_with_info = OrderedDict()
+        for typ in types:
             type_response = OrderedDict()
-            for rid in type_list:
+            for rid in listing[typ]:
                 type_response[rid] = responses[response_index]
                 response_index += 1
-            listing_with_info.append(type_response)
+            listing_with_info[typ] = type_response
 
         return listing_with_info
 
     def info(self,
              cik,
              rid={'alias': ''},
-             #options={"usage": False, "counts": False},
              options={},
              cikonly=False,
-             recursive=False):
+             recursive=False,
+             level=None):
         '''Returns info for RID as a dict.'''
         if recursive:
             rid = None if type(rid) is dict else rid
-            response = self._infotree(cik, rid=rid, options=options)
+            response = self._infotree(cik,
+                                      rid=rid,
+                                      options=options,
+                                      level=level)
         else:
             isok, response = self.exo.info(cik, rid, options)
             self._raise_for_response(isok, response)
@@ -743,11 +795,15 @@ class ExoRPC():
         else:
             return response
 
-    def flush(self, cik, rids):
-        for rid in rids:
-            self.exo.flush(cik, rid, defer=True)
-        if self.exo.has_deferred(cik):
-            self._raise_for_deferred(self.exo.send_deferred(cik))
+    def flush(self, cik, rids, newerthan=None, olderthan=None):
+        args=[]
+        options = {}
+        if newerthan is not None: options['newerthan'] = newerthan
+        if olderthan is not None: options['olderthan'] = olderthan
+        if len(options) > 0:
+            args.append(options)
+        cmds = [['flush', rid] + args for rid in rids]
+        self._exomult(cik, cmds)
 
     def usage(self, cik, rid, metrics, start, end):
         for metric in metrics:
@@ -783,6 +839,11 @@ class ExoRPC():
         self._raise_for_response(isok, response)
         return response
 
+    def clone(self, cik, options):
+        isok, response = self.exo.create(cik, 'clone', options)
+        self._raise_for_response(isok, response)
+        return response
+
     #def tag(self, cik, rids, addtags=[], removetags=[]):
     #    cmds = []
     #    for c in itertools.chain(*[[['map', 'tag', rid, t]
@@ -795,7 +856,13 @@ class ExoRPC():
     #    pprint(cmds)
     #    self._exomult(cik, cmds)
 
-    def _print_node(self, rid, info, aliases, cli_args, spacer, islast, max_name):
+    def _print_tree_line(self, line):
+        if sys.version_info < (3, 0):
+            print(line.encode('utf-8'))
+        else:
+            print(line)
+
+    def _print_node(self, rid, info, aliases, cli_args, spacer, islast, maxlen=None):
         typ = info['basic']['type']
         if typ == 'client':
             id = 'cik: ' + info['key']
@@ -837,7 +904,10 @@ class ExoRPC():
 
         has_alias = aliases is not None and len(aliases) > 0
         if has_alias:
-            add_opt(True, 'aliases', str(aliases))
+            if type(aliases) is list:
+                add_opt(True, 'aliases', json.dumps(aliases))
+            else:
+                add_opt(True, 'aliases', aliases)
         # show RID for clients with no alias, or if --verbose was passed
         ridopt = False
         if typ == 'client':
@@ -848,15 +918,26 @@ class ExoRPC():
         add_opt(ridopt, 'rid', rid)
         add_opt('--verbose', 'unit', units)
 
-        if 'format' in info['description']:
-            desc = info['description']['format'] + ' ' + typ + ' ' + id
-        else:
-            desc = typ + ' ' + id
 
-        print('{0}{1}{2} {3} {4}'.format(
+        if maxlen == None:
+            maxlen = {}
+            maxlen['type'] = len(typ)
+            maxlen['name'] = len(name)
+            maxlen['format'] = 0 if 'format' not in info['description'] else len(info['description']['format'])
+
+        if 'format' in info['description']:
+            fmt = info['description']['format']
+            desc = fmt + ' ' * (maxlen['format'] + 1 - len(fmt))
+            desc += typ + ' ' * (maxlen['type'] + 1 - len(typ))
+            desc += id
+        else:
+            desc = typ + ' ' * (maxlen['type'] + 1 - len(typ))
+            desc += id
+
+        self._print_tree_line('{0}{1}{2} {3} {4}'.format(
             spacer,
             name,
-            ' ' * (max_name - len(name)),
+            ' ' * (maxlen['name'] + 1 - len(name)),
             desc,
             '' if len(opt) == 0 else '({0})'.format(', '.join(
                 ['{0}: {1}'.format(k, v) for k, v in iteritems(opt)]))))
@@ -876,73 +957,94 @@ class ExoRPC():
             # info doesn't contain key
             info['key'] = cik
             aliases = info['aliases']
-            root_aliases = '(see parent)'
-            # todo: can I get aliases for cik? For now, pass []
+            root_aliases = 'see parent'
             self._print_node(rid,
                              info,
                              root_aliases,
                              cli_args,
                              spacer,
-                             True,
-                             len(info['description']['name']))
+                             True)
             if max_level == 0:
                 return
             level += 1
 
         types = ['dataport', 'datarule', 'dispatch', 'client']
         try:
-            listing = self._listing_with_info(cik, types=types, options=info_options)
-            # listing(): [['<rid0>', '<rid1>'], ['<rid2>'], [], ['<rid3>']]
-            # _listing_with_info(): [{'<rid0>':<info0>, '<rid1>':<info1>},
-            #                       {'<rid2>':<info2>}, [], {'<rid3>': <info3>}]
+            listing = self._listing_with_info(cik, types=types, info_options=info_options)
+            # _listing_with_info(): {'client': {'<rid0>':<info0>, '<rid1>':<info1>},
+            #                        'dataport': {'<rid2>':<info2>}}
         except pyonep.exceptions.OnePlatformException:
-            print(spacer + "  └─listing for {0} failed. info['basic']['status'] is probably not valid.".format(cik))
+            self._print_tree_line(
+                spacer +
+                "  └─listing for {0} failed. info['basic']['status'] is \
+probably not valid.".format(cik))
         else:
-            # calculate the maximum length name of all children
-            lengths = [len(l[1]['description']['name']) for i in range(len(types)) for l in iteritems(listing[i])]
-            max_name = 0 if len(lengths) == 0 else max(lengths)
+            # calculate the maximum length of various things for all children,
+            # so we can make things line up in the output.
+            maxlen = {}
+            namelengths = [len(l[1]['description']['name']) for typ in types for l in iteritems(listing[typ])]
+            maxlen['name'] = 0 if len(namelengths) == 0 else max(namelengths)
+
+            typelengths = [len(l[1]['basic']['type']) for typ in types for l in iteritems(listing[typ])]
+            maxlen['type'] = 0 if len(typelengths) == 0 else max(typelengths)
+
+            formatlengths = [len(l[1]['description']['format'])
+                             for typ in types
+                             for l in iteritems(listing[typ])
+                             if 'format' in l[1]['description']]
+            maxlen['format'] = 0 if len(formatlengths) == 0 else max(formatlengths)
 
             # print everything
             for t_idx, t in enumerate(types):
-                typelisting = OrderedDict(sorted(iteritems(listing[t_idx]), key=lambda x: x[1]['description']['name'].lower()))
-                islast_nonempty_type = (t_idx == len(types) - 1) or (all(len(x) == 0 for x in listing[t_idx + 1:]))
+                typelisting = OrderedDict(sorted(iteritems(listing[t]), key=lambda x: x[1]['description']['name'].lower()))
+                islast_nonempty_type = (t_idx == len(types) - 1) or (all(len(listing[typ]) == 0 for typ in types[t_idx + 1:]))
                 for rid_idx, rid in enumerate(typelisting):
                     info = typelisting[rid]
                     islastoftype = rid_idx == len(typelisting) - 1
                     islast = islast_nonempty_type and islastoftype
-                    if islast:
-                        child_spacer = spacer + '    '
-                        own_spacer   = spacer + '  └─'
+                    if platform.system() != 'Windows':
+                        if islast:
+                            child_spacer = spacer + '    '
+                            own_spacer   = spacer + '  └─'
+                        else:
+                            child_spacer = spacer + '  │ '
+                            own_spacer   = spacer + '  ├─'
                     else:
-                        child_spacer = spacer + '  │ '
-                        own_spacer   = spacer + '  ├─'
+                        # Windows executable
+                        if islast:
+                            child_spacer = spacer + '    '
+                            own_spacer   = spacer + '  +-'
+                        else:
+                            child_spacer = spacer + '  | '
+                            own_spacer   = spacer + '  +-'
 
                     if t == 'client':
                         next_cik = info['key']
-                        self._print_node(rid, info, aliases, cli_args, own_spacer, islast, max_name)
+                        self._print_node(rid, info, aliases, cli_args, own_spacer, islast, maxlen)
                         if max_level == -1 or level < max_level:
                             self.tree(next_cik, info['aliases'], cli_args, child_spacer, level=level + 1, info_options=info_options)
                     else:
-                        self._print_node(rid, info, aliases, cli_args, own_spacer, islast, max_name)
+                        self._print_node(rid, info, aliases, cli_args, own_spacer, islast, maxlen)
 
     def drop_all_children(self, cik):
-        isok, listing = self.exo.listing(cik,
-            types=['client', 'dataport', 'datarule', 'dispatch'])
+        isok, listing = self.exo.listing(
+            cik,
+            types=['client', 'dataport', 'datarule', 'dispatch'],
+            options={})
         self._raise_for_response(isok, listing)
-
-        for l in listing:
-            self.drop(cik, l)
+        rids = itertools.chain(*[listing[t] for t in listing.keys()])
+        self._exomult(cik, [['drop', rid] for rid in rids])
 
     def _lookup_rid_by_name(self, cik, name, types=['datarule']):
         '''Look up RID by name. We use name rather than alias to identify
-        scripts created in Portals, which only shows names, not aliases.
-        Note that if multiple scripts have the same name, the first one
-        in the listing is returned.'''
+        scripts created in Portals because it only displays names to the
+        user, not aliases. Note that if multiple scripts have the same
+        name the first one in the listing is returned.'''
         found_rid = None
         listing = self._listing_with_info(cik, types)
-        for type_listing in listing:
-            for rid in type_listing:
-                if type_listing[rid]['description']['name'] == name:
+        for typ in listing:
+            for rid in listing[typ]:
+                if listing[typ][rid]['description']['name'] == name:
                     # return first match
                     return rid
         return None
@@ -987,10 +1089,10 @@ class ExoRPC():
         fn(cik)
         lwi = self._listing_with_info(cik,
                                       ['client'],
-                                      {'key': True})
-        # [{'<rid0>':<info0>, '<rid1>':<info1>}]
-        for rid in lwi[0]:
-            self.cik_recursive(lwi[0][rid]['key'], fn)
+                                      info_options={'key': True})
+        # {'client': {'<rid0>':<info0>, '<rid1>':<info1>}]
+        for rid in lwi['client']:
+            self.cik_recursive(lwi['client'][rid]['key'], fn)
 
     def upload_script(self,
                       ciks,
@@ -1026,10 +1128,10 @@ class ExoRPC():
                         up(cik)
 
     def lookup_rid(self, cik, cik_to_find):
-        isok, listing = self.exo.listing(cik, types=['client'])
+        isok, listing = self.exo.listing(cik, types=['client'], options={})
         self._raise_for_response(isok, listing)
 
-        for rid in listing[0]:
+        for rid in listing['client']:
             self.exo.info(cik, rid, {'key': True}, defer=True)
 
         if self.exo.has_deferred(cik):
@@ -1039,7 +1141,7 @@ class ExoRPC():
                 self._raise_for_response(isok, response)
 
                 if response['key'] == cik_to_find:
-                    return listing[0][idx]
+                    return listing['client'][idx]
 
     def record_backdate(self, cik, rid, interval_seconds, values):
         '''Record a list of values and record them as if they happened in
@@ -1100,10 +1202,16 @@ class ExoRPC():
         '''Make a copy of cik and its non-client children to destcik and
         return the cik of the copy.'''
 
+
         # read in the whole client to copy at once
         if infotree is None:
+            def check_for_unsupported(rid, info):
+                desc = info['description']
+                if 'subscribe' in desc and desc['subscribe'] is not None and len(desc['subscribe']) > 0:
+                    raise ExoException('''Copy does not yet support resources that use the "subscribe" feature, as RID {0} in the source client does.\nIf you're just copying a device into the same portal consider using the clone command.'''.format(rid));
+                return rid
             destcik = self.lookup_shortcut(destcik)
-            infotree = self._infotree(cik, options={})
+            infotree = self._infotree(cik, options={}, nodeidfn=check_for_unsupported)
 
         # check counts
         counts = self._counttypes(infotree)
@@ -1159,7 +1267,7 @@ class ExoRPC():
 
         return list(differ.compare(s1, s2))
 
-    def _infotree(self, cik, rid=None, nodeidfn=lambda rid, info: rid, options={}):
+    def _infotree(self, cik, rid=None, nodeidfn=lambda rid, info: rid, options={}, level=None):
         '''Get all info for a cik and its children in a nested dict.
         The basic unit is {'rid': '<rid>', 'info': <info-with-children>},
         where <info-with-children> is just the info object for that node
@@ -1184,27 +1292,36 @@ class ExoRPC():
         types = ['client', 'dataport', 'datarule', 'dispatch']
         #print(cik, rid)
         # TODO: make exactly one HTTP request per node
-        listing = []
+        listing = {}
         norid = rid is None
         if norid:
             rid = self._exomult(cik, [['lookup', 'aliased', '']])[0]
 
         info = self._exomult(cik, [['info', rid, options]])[0]
 
+        myid = nodeidfn(rid, info)
+
+        if level is not None and level <= 0:
+            return {'rid': myid, 'info': info}
+
         if norid or info['basic']['type'] == 'client':
             if not norid:
                 # key is only available to owner (not the resource itself)
                 cik = info['key']
             listing = self._exomult(cik,
-                                    [['listing', types]])[0]
-
-        myid = nodeidfn(rid, info)
+                                    [['listing', types, {}]])[0]
 
         info['children'] = []
-        for typ, ridlist in zip(types, listing):
-            for childrid in ridlist:
-                tr = self._infotree(cik, childrid, nodeidfn=nodeidfn, options=options)
-                info['children'].append(tr)
+        for typ in types:
+            if typ in listing:
+                ridlist = listing[typ]
+                for childrid in ridlist:
+                    tr = self._infotree(cik,
+                                        childrid,
+                                        nodeidfn=nodeidfn,
+                                        options=options,
+                                        level=None if level is None else level-1)
+                    info['children'].append(tr)
         info['children'].sort(key=itemgetter('rid'))
 
         return {'rid': myid, 'info': info}
@@ -1320,7 +1437,7 @@ class ExoRPC():
 
 class ExoData():
     '''Implements the Data Interface API
-    https://github.com/exosite/api/tree/master/data'''
+    https://github.com/exosite/docs/tree/master/data'''
 
     def __init__(self, url='http://m2.exosite.com'):
         self.url = url
@@ -1364,7 +1481,7 @@ class ExoData():
 
 class ExoProvision():
     '''Implements the Provision API
-    https://github.com/exosite/api/tree/master/provision'''
+    https://github.com/exosite/docs/tree/master/provision'''
 
     def __init__(self, url='http://m2.exosite.com'):
         self.url = url
@@ -1418,26 +1535,24 @@ class ExoPortals():
 
 
 def parse_ts(s):
-    return parse_ts_tuple(parser.parse(s).timetuple())
+    return None if s is None else parse_ts_tuple(parser.parse(s).timetuple())
 
 def parse_ts_tuple(t):
     return int(time.mktime(t))
 
 def is_ts(s):
-    return re.match('^[0-9]+$', s) is not None
+    return s is not None and re.match('^[0-9]+$', s) is not None
 
 def get_startend(args):
     '''Get start and end timestamps based on standard arguments'''
     start = args.get('--start', None)
     end = args.get('--end', None)
-    if start is None:
-        start = 1
-    elif is_ts(start):
+    if is_ts(start):
         start = int(start)
     else:
         start = parse_ts(start)
-    if end is None or end == 'now':
-        end = int(parse_ts_tuple(datetime.now().timetuple()))
+    if end == 'now':
+        end = None
     elif is_ts(end):
         end = int(end)
     else:
@@ -1581,8 +1696,8 @@ def read_cmd(er, cik, rids, args):
     if len(rids) == 0:
         # if only a CIK was passed, include all dataports and datarules
         # by default.
-        listing = er.listing(cik, ['dataport', 'datarule'])
-        rids = [rid for rid in itertools.chain(*listing)]
+        listing = er.listing(cik, ['dataport', 'datarule'], {})
+        rids = listing['dataport'] + listing['datarule']
         aliases = er.info(cik, options={'aliases': True})['aliases']
         # look up aliases for column headers
         cmdline_rids = [aliases[rid][0] if rid in aliases else rid for rid in rids]
@@ -1729,7 +1844,8 @@ def handle_args(cmd, args):
                 port=args['--port'],
                 https=use_https,
                 httptimeout=args['--httptimeout'],
-                logrequests=args['--clearcache'])
+                logrequests=args['--clearcache'],
+                user_agent=args['--useragent'])
     if cmd in ['ip', 'data']:
         if args['--https'] is True or args['--port'] is not None or args['--debughttp'] is True:
             # TODO: support these
@@ -1933,18 +2049,22 @@ def handle_args(cmd, args):
                 for key in exclude.split(',')]
 
             options = er.make_info_options(include, exclude)
+            level = args['--level']
+            level = None if level is False or args['--recursive'] is False else int(level)
             info = er.info(cik,
                         rids[0],
                         options=options,
                         cikonly=args['--cikonly'],
-                        recursive=args['--recursive'])
+                        recursive=args['--recursive'],
+                        level=level)
             if args['--pretty']:
                 pr(info)
             else:
                 # output json
                 pr(json.dumps(info))
         elif cmd == 'flush':
-            er.flush(cik, rids)
+            start, end = get_startend(args)
+            er.flush(cik, rids, newerthan=start, olderthan=end)
         elif cmd == 'usage':
             allmetrics = ['client',
                         'dataport',
@@ -2071,6 +2191,32 @@ def handle_args(cmd, args):
                 typ = 'client'
                 code = args['--client']
             er.deactivate(cik, typ, code)
+        elif cmd == 'clone':
+            options = {}
+            if args['--share'] is not None:
+                options['code'] = args['--share']
+            if args['--rid'] is not None:
+                rid_to_clone = args['--rid']
+                if er.regex_rid.match(rid_to_clone) is None:
+                    # try to look up RID for an alias
+                    alias = rid_to_clone
+                    rid_to_clone = er.lookup(cik, alias)
+                options['rid'] = rid_to_clone
+
+            options['noaliases'] = args['--noaliases']
+            options['nohistorical'] = args['--nohistorical']
+
+            rid = er.clone(cik, options)
+            pr('rid: {0}'.format(rid))
+            info = er.info(cik, rid, {'basic': True, 'key': True})
+            typ = info['basic']['type']
+            copycik = info['key']
+            if typ == 'client':
+                if not args['--noactivate']:
+                    er.activate(cik, 'client', copycik)
+                # for convenience, look up the cik
+                pr('cik: {0}'.format(copycik))
+
         #elif cmd == 'tag':
             # One Platform does not yet support removing tags.
             #removetags = args['--remove']
@@ -2110,11 +2256,15 @@ class DiscreetFilter(object):
        an RID'''
     def __init__(self, out):
         self.out = out
+        # match the two halves of an RID/CIK
         self.ridre = re.compile('([a-fA-F0-9]{20})([a-fA-F0-9]{20})')
 
     def write(self, message):
-        self.out.write(self.ridre.sub(r'\g<1>01234567890123456789',
-                                      message))
+        # hide the second half
+        message = message.decode('utf-8')
+        s = self.ridre.sub('\g<1>01234567890123456789', message)
+        s = s.encode('utf-8')
+        self.out.write(s)
 
 def cmd(argv=None, stdin=None, stdout=None, stderr=None):
     '''Wrap the command line interface. Globally redirects args
